#include "bezier.h"

// Default constructor.  It creates a "curve" with 0 control points.  The only real duties of this function
// are to give everything default values, allocate a single basis matrix, and give it its values.
bezier::bezier(): cont(G0)
{	GLfloat mat[16]  =  {  -1, 3,-3, 1 ,				// Bezier basis matrix
					3,-6, 3, 0,
					-3, 3, 0, 0,
					1, 0, 0, 0 };
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	showPoints = true;
	detail = 50;
	number = 0;
	dirty = true;
	points = NULL;
	for(int i = 0; i < 3; i++)
		coeffs[i] = NULL;
}

// This constructor sets up the appropriate number of points, but they are all currently
// on the origin.  The curve won't be very interesting until they are changed.
bezier::bezier(unsigned int num): cont(G0)
{	GLfloat mat[16]  =  {  -1, 3,-3, 1 ,				// Bezier basis matrix
					 3,-6, 3, 0 ,
					-3, 3, 0, 0 ,
					 1, 0, 0, 0 };
	number = num;
	
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	if(num > 0)								// Now allocate space to store the
	{	points = new coordinate[number];			// control points and coefficients
		if(num > 3)							// (if necessary).
			for(int i = 0; i < 3; i++)
				coeffs[i] = new coefficient[(number-1)/3];
		else coeffs[0] = coeffs[1] = coeffs[2] = NULL;
	}
	else points = NULL;
	
	for(int i = 0; i < number; i++)					// Initialize all points to origin
		for(int j = 0; j < 3; j++)
			points[i][j] = 0;
	
	showPoints = true;
	detail = 50;
	dirty = true;
}

// This constructor sets up a curve with the specified number of points whose values are
// provided by the array.
bezier::bezier(unsigned int num, GLfloat * array): cont(G0)
{	GLfloat mat[16]  =  {  -1, 3,-3, 1 ,				// Bezier basis matrix
					 3,-6, 3, 0 ,
					-3, 3, 0, 0 ,
					 1, 0, 0, 0 };
	
	number = num;
	M = new basis[1];							// Allocate basis matrix
	M[0] = mat;								// and give it values.
	
	if(num > 0)								// Now allocate space to store the
	{	points = new coordinate[number];			// control points and coefficients
		if(num > 3)							// (if necessary).
			for(int i = 0; i < 3; i++)
				coeffs[i] = new coefficient[(number-1)/3];
		else coeffs[0] = coeffs[1] = coeffs[2] = NULL;
	}
	else points = NULL;
	
	for(int i = 0; i < number; i++)					// Initialize all points to values passed
		for(int j = 0; j < 3; j++)					// in the array
			points[i][j] = array[i*3+j];
	
	showPoints = true;
	detail = 50;
	dirty = true;
}

// Returns the continuity of the curve.
Continuity bezier::getContinuity() const
{	return cont;
}

// This returns the number of segments in the complete cuve.  The first four points define one
// segment, and because of the shared control point, every additional three create one more.
// So just subtract 1 from the number of points and divide it by 3.
int bezier::howManySegments() const
{	return (number - 1) / 3;
}

// Because of special considerations about continuity, the bezier class must create its own functions
// for setting the values of the control points.
// The logic behind the set*() functions is the same as that of the general curve class, but if the
// continuity is G1 or C1, then we also must make the proper call to enforceCont().
void bezier::set(unsigned int index, coordEnum which, GLfloat value)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to bezier::set().\n";
		return;
	}
	
	points[index][which] = value;
	
	dirty = true;							// Mark the curve as changed.
	
	if(cont == G0 || index < 2)					// End without worrying about continuity
		return;							// if it is G0 or the start of the curve.
	switch (index%3)
	{	case 0:							// This is a join point.  Set the one to the left
			if(index < number - 1)			// IF a point to the right exists.
				enforceCont(index, -1);
			break;
		case 1:							// This is to the "right" of the joint, so set the
			enforceCont(index-1, -1);			// left one.
			break;
		case 2:							// index is to the "left" of the joint, so set the
			if(index < number - 2)			// "right" point IF there one exists.
				enforceCont(index+1, 1);
			break;
	}
}

// Changes the values of control point #index to those passed in the array.  An error will be
// printed and nothing changed if index is not valid.
void bezier::set(unsigned int index, GLfloat *array)
{	if(index >= number)
	{	cerr << "Error.  Illegal value passed to bezier::set().\n";
		return;
	}
	
	for(int i = 0; i < 3; i++)
		points[index][i] = array[i];
	
	dirty = true;							// Mark the curve as changed.
	
	if(cont == G0 || index < 2)					// End without worrying about continuity
		return;							// if it is G0 or the start of the curve.
	switch (index%3)
	{	case 0:							// This is a join point.  Set the one to the right
			enforceCont(index, 1);
				if(index < number - 1)		// IF a point to the right exists.
			break;
		case 1:							// This is to the "right" of the joint, so set the
			enforceCont(index-1, -1);			// left one.
			break;
		case 2:							// index is to the "left" of the joint, so set the
			if(index < number - 2)			// "right" point IF one exists there.
				enforceCont(index+1, 1);
			break;
	}
}

// This function sets all the control points to the values in the array.  The total number of control
// points remains unchanged.  No error checking is done to make sure that the user has
// provided enough values -- Everything accessed is considered to be a valid GLfloat.
// If the curve is C1 or G1 continuous, then the continuity is preserved by ignoring the values to the
// "left" of the internal join points.  Yes, this ignores some of the user's input, but continuity is
// necessarily restrictive.
void bezier::setAll(GLfloat * array)
{	for(int i = 0; i <number; i++)
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	if(cont != G0)							// Do we need to make the curve continuous?
		for(int i = 3; i < number - 1; i++)			// For each of the internal join points...
			enforceCont(i, -1);				// alter the points to be continuous.
	dirty = true;							// The coefficients need to be recalculated
}

// This function changes the number of control points to newnum and gives them the values
// passed in the array.  It will also, if necessary, reallocate memory for coeffs.  (Remember
// that every segment is made up of four points with a single one shared between segments.)
// If the curve is C1 or G1 continuous, then the continuity is preserved by changing the values to the
// "left" of the internal join points.  This ignores some of the user's input, but continuity is
// necessarily restrictive.
void bezier::setAll(unsigned int newnum, GLfloat * array)
{	if(newnum != number)
	{	if(number > 0)						// If there are already points, we first need to
		{	delete [] points;					// deallocate the memory of them.  Don't worry
			for(int i = 0; i < 3; i++)			// about saving it...everything will change.
				delete [] coeffs[i];	
		}
		if(newnum > 0)						// If there are going to be any points, we need
		{	points = new coordinate[newnum];	// to allocate memory for them.
			if(newnum >=3)				// And if there are enough points for a curve
				for(int i = 0; i < 3; i++)		// segment, allocate space for them, too.
					coeffs[i] = new coefficient[(newnum-1) / 3];
		}
		number = newnum;
	}
	
	// Whatever happened above, we know that points and number are now correct.
	for(int i = 0; i <number; i++)				// Copy values of control points
		for(int j = 0; j < 3; j++)
			points[i][j] = array[i*3 + j];
	if(cont != G0)							// Do we need to make the curve continuous?
		for(int i = 3; i < number - 1; i++)			// For each of the internal join points...
			enforceCont(i, -1);				// alter the points to be continuous.
	dirty = true;							// The coefficients need to be recalculated
}

// This function adds a control point to the curve.  If we now have enough control points to make a
//  new segment, do so.  If the control point is the second one (first after the shared endpoint) in the
// curve, then we may need to position it so that the continuity is preserved.  In this case, it is put on
// the origin and then made continuous.  Be aware that this may cause undesirable distances with
// G1 continuity.
// If the curve is G0 continuous, then the new point is merely placed on the origin.
void bezier::addPoint()
{	if(number > 0)							// Save the points that already exist.
	{	coordinate * temp = new coordinate[number+1];
		for(int i = 0; i < number; i++)
			for(int j = 0; j < 3; j++)
				temp[i][j] = points[i][j];
		
		delete [] points;
		points = temp;
	}
	else									// Otherwise, we just need room for one new
		points = new coordinate[1];			// point.
	
	
	number++;							// Increase our count of control points.
	
	for(int i = 0; i < 3; i++)					// Now set the point's location (the origin).
		points[number-1][i] = 0;
	if((number-1)%3 == 1 && number > 4)		// If this is the second point in a segment,
		enforceCont(number-2, 1);			// make sure that continuity is preserved.
	
	
	if((number-1)%3 == 0)					// We now have enough for a new segment
	{	if(number > 3)						// First, we need to deallocate the old coefficient
			for(int i = 0; i < 3; i++)			// info if it exists (ie, if this isn't the first segment.)
				delete [] coeffs[i];			// Then we allocate new memory for them and
		for(int i = 0; i < 3; i++)				// calculate their values.  (Seperate loops are
			coeffs[i] = new coefficient[number/3];// used so that as much memory as possible
										// is freed before we allocate more.)
		dirty = true;						// The coefficients need to be recalculated
	}
}

// This function adds a control point to the curve, with a location given by the array.  If we now have
// enough control points to make a new segment, do so.  Because the location specified by the
// user may be important, we keep it if we need to change the continuity.  This is different from the
// way that continuity was handled in the previous addPoint() function, but the change is necessary
// due to the differences in the information we are given.
void bezier::addPoint(GLfloat * array)
{	if(number > 0)							// Save the points that already exist.
	{	coordinate * temp = new coordinate[number+1];
		for(int i = 0; i < number; i++)
			for(int j = 0; j < 3; j++)
				temp[i][j] = points[i][j];
		
		delete [] points;
		points = temp;
	}
	else									// Otherwise, we just need room for one new
		points = new coordinate[1];			// point.
	
	
	number++;							// Increase our count of control points.
	for(int i = 0; i < 3; i++)					// The new control point defaults to the origin
		points[number-1][i] = array[i];
	
	if((number-1)%3 == 1 && number > 4 && cont != G0)
	{	enforceCont(number-2, -1);			// Must retain continuity
		dirty = true;						// This changes the shape of a segment
	}
	if((number-1)%3 == 0)					// We now have enough for a new segment
	{	if(number > 3)						// First, we need to deallocate the old coefficient
			for(int i = 0; i < 3; i++)			// info if it exists (ie, if this isn't the first segment.)
				delete [] coeffs[i];			// Then we allocate new memory for them
		for(int i = 0; i < 3; i++)				// and calculate their values.
			coeffs[i] = new coefficient[(number-1)/3];
		dirty = true;						// The coefficients need to be recalculated
	}
}

// This function changes the curve's continuity of its join points to the requested degree.
void bezier::setContinuity(Continuity arg)
{	if(arg == cont)						// If nothing's being changed, leave the function
		return;						// without going through all the trouble below.
	switch(arg)
	{	case G0:						// If G0, we don't need to change anything.  So the
			cont = G0;					// points aren't moved and the dirty flag isn't set.
			break;
		case G1:						// For G1 and C1 continuity, we move the points
			cont = G1;					// to the "right" of the join points so that we have the
			for(int i = 3; i < number-1; i+=3)// needed continuity.  These for() loops iterate once
				enforceCont(i, -1);		// for each internal joint (3, 6, 9, . . . as long as there
			dirty = true;				// are at least 3 more control points to the right.  This
			break;					// end condition makes sure that there is a segment
		case C1:						// on both sides of the joint.)  Then it is simply
			cont = C1;					// necessary to call the enforceCont() function, which
			for(int i = 3; i < number-1; i+=3)// moves the control point for us.
				enforceCont(i, -1);
			dirty = true;
			break;
	}
}

// Calculate the value of the equation for a specific value of t, and return the value of the desired
// coordinate (either X, Y, or Z).  If t < 0, then it shall be assumed that the first segment stretches
// to -.  If t > (number-1)/3 then it shall be assumed that the last segment goes to +.  If there
// is no valid segment, 0 is returned.
GLfloat bezier::getAtTimeT(coordEnum which, GLfloat t)
{	int segment = t;					// integer value of t is which segment we use.
	GLfloat localT = t - segment;			// Value of t within segment, normalized to [0..1]
	GLfloat t2, t3;						// localT^2 and localT^3, with easier names.
	
	if (number < 4)						// If there aren't any segments yet, return 0
		return 0;
	
	if(segment < 0)						// Can't be before the 0th segment
	{	segment = 0;
		localT = t;
	}
	else if(segment > (number-1)/3)		// Can't be beyond last segment
	{	segment = (number-1)/3;
		localT = t - segment;
	}
	
	if(dirty)							// Do we need to recalculate the coefficients?
		calcCoeffs();
	
	t2 = localT*localT;
	t3 = t2 * localT;
	
	return (coeffs[which][segment][0]*t3 + coeffs[which][segment][1]*t2 + 
		    coeffs[which][segment][2]*localT + coeffs[which][segment][3]);
}

// Calculate the value of the equation for a specific value of t.  The answer is put in the array that
// has been passed to the function.  If t < 0, then it shall be assumed that the first segment
// stretches to -.  If t > (number-1)/3 then it shall be assumed that the last segment goes to +.
// If there is no valid segment, (0, 0, 0) is returned.
void bezier::getAtTimeT(GLfloat t, GLfloat * array)
{	int segment = t;					// integer value of t is which segment we use.
	GLfloat localT = t - segment;			// Value of t within segment, normalized to [0..1]
	GLfloat t2, t3;						// localT^2 and localT^3, with easier names.
	
	if (number < 4)						// If there aren't any segments yet, return (0, 0, 0)
		for(int i = 0; i < 3; i++)
			array[i] = 0;
	
	if(segment < 0)						// Can't be before the 0th segment
	{	segment = 0;
		localT = t;
	}
	else if(segment >= (number-1)/3)		// Can't be beyond last segment
	{	segment = (number-1)/3 - 1;
		localT = t - segment;
	}
	
	if(dirty)							// Do we need to recalculate the coefficients?
		calcCoeffs();
	
	t2 = localT*localT;
	t3 = t2 * localT;
	
	for(int i = 0; i < 3; i++)				// Calculate for the given t
		array[i] = coeffs[i][segment][0]*t3 + coeffs[i][segment][1]*t2 + 
			        coeffs[i][segment][2]*localT + coeffs[i][segment][3];
}

// Draw the curve and possibly its control points
void bezier::draw()
{	if(dirty)							// If the values of the curve have been changed,
		calcCoeffs();					// then calculate the new coefficients.
	
	if(showPoints)						// If we are showing the control points,
	{	glPointSize(5.0);				// then do so.
		glColor3f(1.0, 1.0, 1.0);			// Set a color to white, point size to 5
		glBegin(GL_POINTS);
			for(int i = 0; i < number; i++)
				glVertex3fv(points[i]);
		glEnd();
		glColor3f(0.5, 0.5, 0.7);			// Set bluish-gray color for drawing lines between
		glEnable(GL_LINE_STIPPLE);		// P0/P1 and P2/P3.  Use stippling to make them
		glLineStipple(1, 0x9248);		// dashed.
		glBegin(GL_LINE_STRIP);		// Tell OpenGL to start making a line strip.
			for(int i = 0; i < number;i++)	// Specify vertices one by one
				glVertex3fv(points[i]);
		glEnd();						// Tell OpenGL to stop making the line strip.
		glDisable(GL_LINE_STIPPLE);
	}
	
	glColor3f(1.0, 1.0, 1.0);				// Set the color to white
	for(int i = 0; i < (number-1)/3; i++)		// Now draw the curve segments themselves.
		cubicCurve(&coeffs[0][i][0], &coeffs[1][i][0], &coeffs[2][i][0], detail);
}

// This function is used to keep the curve continuous.  When a change is made that could break
// the continuity, this function should be called.  The first parameter indexes the start/endpoint of
// a segment around which the continuity is threatened.  To preserve continuity, one of the two
// points "next" to that control point must be moved.  The second parameter, which must be
// either 1 or -1 (because this is a protected function, the assumption is that the arguments are
// valid) tells which one this is.  Point #(index+change) will be changed so that it meets the
// requirements set by point #index and point #(index-change).
void bezier::enforceCont(unsigned int index, int change)
{	GLfloat dist1, dist2;					// These will hold distances between points
	switch(cont)						// How we do this depends on the degree of
	{	case G0:						// continuity we are requiring.
			break;					// G0 (or C0) requires no effort.
		// For G1 continuity, the three points determining the shape of the curve at the joint
		// must be colinear with both points #(index+1) and #(index-1) being on the same
		// side of point #index.  However, we can preserve the distances to point #index.
		case G1:	
			dist1 = dist(points[index], points[index-change]);	// Dist to point that won't move
			dist2 = dist(points[index], points[index+change]);	// Preserve this distance
			for(int i = 0; i < 3; i++)						// while moving the point
				points[index+change][i] =
					points[index][i]-(points[index-change][i]-points[index][i])*(dist2/dist1);
			break;
		// C1 continuity is easy.  The points on both "sides" of the joint must have the same
		// location.
		case C1:
			for(int i = 0; i < 3; i++)
				points[index+change][i] = 2*points[index][i] - points[index-change][i];
			break;
	}
	dirty = true;
}