/* This example program demonstrates basic Bzier curves.  Control-clicking adds new control
 * points, and dragging on the screen will move the curve around.  The option to watch an
 * object move along the curve over time is cool.
 */

#include "bezier.h"

// #defined values to be used when showing the object moving along the curve.
#define T_SPEED .01								// Step between successive t values
#define PAUSE_TIME 20							// Milliseconds before t changes
#define STEPS_BETWEEN_SAVE 10					// Every this-many-number of times,
												// record the value of t.
// #define values for menu options
#define MENU_G0 1
#define MENU_G1 2
#define MENU_C1 3
#define MENU_MOVEMENT 10
#define MENU_QUIT 911

// Prototypes
void display();
void init();
void reshape(int, int);
void keyboard(unsigned char, int, int);
void click(int, int, int, int);
void drag(int, int);
void menu(int);
void showMovie(int);

int dragPoint;			// Which point is currently being dragged.. Default is -1
int windowHeight;		// Based on w and h parameters to reshape()

GLfloat points[12] = { 25.0, 25.0, 0.0,			// Values for the four coordinates
					   100.0, 100.0, 0.0,
					   175.0, 175.0, 0.0,
					   90.0, 190.0, 0.0 };
int winLeft, winBottom;	// What the coordinate should be on the lower left of the window
int dragX, dragY;		// Used for dragging the whole curve. (See click() and drag().)
coordinate * tSaved;	// Values at specific times of t.  Used for the "movie".
bool showingMovie;		// Is it currently showing the change in t over time?

bezier bcurve(4, points);

void display(void)
{	glClear(GL_COLOR_BUFFER_BIT);
	
	bcurve.draw();
	
	glutSwapBuffers ();
}

void init (void) 
{	glClearColor (0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	
	dragPoint = -1;
	winLeft = winBottom = 0;
	showingMovie = false;
}

void reshape (int w, int h)
{	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);	// Switch to projection matrix
	glLoadIdentity ();				// Reset projection matrix
	glOrtho(0, w, 0, h, -1.0, 1.0);	// Set up viewing area
	glMatrixMode (GL_MODELVIEW);	// Return to modelview matrix
	windowHeight = h;				// Remember height of window
}

void keyboard(unsigned char key, int x, int y)
{
#pragma unused (x, y)
	switch (key)
	{	case 27:					// escape key.  Quit.
			bcurve.setAll(0,NULL);	// free memory (since the GLUT won't do it.)
			//if(showingMovie)		// If currently showing the "movie", need to
				//delete [] tSaved;	// delete these values, as well.
			exit(0);
			break;
	}
//	glutPostRedisplay();
}

void click(int button, int state, int x, int y)
{
#pragma unused (button)
	y = windowHeight - y;		// MAC SPECIFIC: Convert from system's coordinate system to ours
	int x2 = x - winLeft;						// Convert from window coords to curve coords
	int y2 = y - winBottom;
	dragPoint = -1;								// Lable point being dragged as default (none)
	int points = bcurve.howManyPoints();		// Number of points in the curve
	if(button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)						// Only do this if the mouse was clicked down
	{	//if(glutGetModifiers() == GLUT_ACTIVE_CTRL)	// If control key down, make new point
			GLfloat newPoint[3] = {x2, y2, 0};
			bcurve.addPoint(newPoint);
			dragPoint = points;
	}	
		else if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)								// Else, find existing point to drag
		{	int minDist = 10;				// Look for closest point within certain tolerance.
			for(int i = 0; i < points; i++)	
				if(dist(x2, y2, bcurve.get(i, X), bcurve.get(i, Y)) < minDist)
				{	minDist = dist(x2, y2, bcurve.get(i, X), bcurve.get(i, Y));
					dragPoint = i;			// Mark closest point as being dragged.
				}
			if(dragPoint == -1)				// If we couldn't find a nearby point
			{	dragX = x;					// Then drag the coordinate system around
				dragY = y;					// (Note that we use windows coords for this.)
			}
		}
	glutPostRedisplay();
}

void drag(int x, int y)
{	y = windowHeight - y;		// Convert from system's coordinate system to ours
	int x2 = x - winLeft;		// x and y have window position, x2 and y2 have the
	int y2 = y - winBottom;		// curve's coordinate system.
	if(dragPoint > -1)
	{	bcurve.set(dragPoint, X, (GLfloat) x2);
		bcurve.set(dragPoint, Y, (GLfloat) y2);
	}
	else						// Drag everything
	{	winLeft += x - dragX;					// Determine new lower-left corner of 
		winBottom += y - dragY;					// the window.
		dragX = x;								// Update dragX and dragY in case the
		dragY = y;								// user keeps dragging.
		// Now update OpenGL stuff:
		glLoadIdentity();						// Reset modelview matrix to default
		glTranslatef(winLeft, winBottom, 0.0);	// and then translate to current offset.
	}
	glutPostRedisplay();
}

void menu(int arg)
{	switch(arg)
	{	case MENU_G0:				// G0/C0 continuity
			bcurve.setContinuity(G0);
			break;
		case MENU_G1:				// G1 continuity
			bcurve.setContinuity(G1);
			break;
		case MENU_C1:				// C1 continuity
			bcurve.setContinuity(C1);
			break;
		case MENU_MOVEMENT:			// Show a point moving along the curve over time
			// First, allocate space to save the coordinate every "STEPS_BETWEEN_SAVE" steps"
			tSaved = new coordinate[(int) 
				(bcurve.howManySegments() / (STEPS_BETWEEN_SAVE * T_SPEED)) + 1];
			bcurve.displayPoints(false);// Hide control points during the movie
			showingMovie = true;	// Mark that we are showing the "movie"
			showMovie(0);			// Then, start moving a point along the curve.
			break;
		case MENU_QUIT:				// Quit program
			//bcurve.setAll(0,NULL);	// free memory (since the GLUT won't do it.)
			//f(showingMovie)		// If currently showing the "movie", need to
				//delete [] tSaved;	// delete these values, as well.
			exit(0);
	}
	glutPostRedisplay();
}

// This function is used when we are showing a point moving along the curve as t varies over
// time.  The parameter tells it what the current time is.  It draws the curve, then puts the
// point over it, and asks the GLUT to call it again in a few milliseconds with a new value
// for t.
// Due to limitations in the GLUT, this function MUST take an int, even though t is a float.
// Therefore, we are using a "step" counter, which is multiplied by the change in t per step
// to get the real value of t.
// (This function has two sets of glBegin(GL_POINTS)..glEnd() calls.  Ideally, these should be
// simplified to one.  For some reason, though, the point size can only be changed outside of
// the glBegin()/glEnd() pair.  I'm assuming this is a bug in Apple's current implementation
// of OpenGL, especially since trying to change the point size there caused portions of the
// curve disappear and made memory deallocation very unstable.)
void showMovie(int step)
{	GLfloat coord[3];
	GLfloat t = step * T_SPEED;			// Calculate current value of t
	
	glClear(GL_COLOR_BUFFER_BIT);		// Clear the window
	bcurve.draw();						// And draw the curve
	
	if(t <= bcurve.howManySegments())	// If it's not done travelling along the curve
	{	bcurve.getAtTimeT(t, coord);	// then get the value at time t
		
		glPointSize(8.0);				// Define a big point size,
		glColor3f(0.9, 0.1, 0.1);		// set a reddish color,
		glBegin(GL_POINTS);				// and draw the point.
			glVertex3fv(coord);
		glEnd();
		
		if(step%STEPS_BETWEEN_SAVE == 0)// Is it time to save a coordinate?
			for(int i = 0; i < 3; i++)
				tSaved[step/STEPS_BETWEEN_SAVE][i] = coord[i];
	}
	
	// Now, draw the points already passed, whether or not this is the end
	glPointSize(4.0);					// Ask for small points
	glColor3f(0.5, 0.5, 0.5);			// Set a grayish color
	glBegin(GL_POINTS);					// Tell OpenGL we're going to draw points
		for(int i = 0; i <= step/STEPS_BETWEEN_SAVE; i++)
			glVertex3fv(tSaved[i]);		// Tell OpenGL to draw the previous points
	glEnd();							// Tell OpenGL that we're done making points
	
	glutSwapBuffers();					// Print the changes out to the screen.
	
	if(t > bcurve.howManySegments())	// If we are now done
	{	delete [] tSaved;				// release memory for the saved coordinates
		bcurve.displayPoints(true);		// and display the control points again.
		showingMovie = false;			// Mark that we are done (for memory purposes)
	}
	else								// Otherwise, ask the GLUT to call this again.
		glutTimerFunc(PAUSE_TIME, showMovie, step+1);
}

// This is a standard main() function for a GLUT-based program.  A single window is opened, 
// callback functions are set, and a menu (which allows the user to set the curve's continuity
// and quit) is attached to the right mouse button.
int main(int argc, char** argv)
{	glutInit(&argc, argv);
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB);
	glutInitWindowSize (250, 250);
	glutInitWindowPosition (100, 100);
	
	
	glutCreateWindow ("Bezier curves");
	init ();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(click);
	glutMotionFunc(drag);
	
	glutCreateMenu(menu);
	glutAddMenuEntry("Allow G0 and C0 continuity", MENU_G0);
	glutAddMenuEntry("Enforce G1 continuity", MENU_G1);
	glutAddMenuEntry("Enforce C1 continuity", MENU_C1);
	glutAddMenuEntry("-", MENU_QUIT);				// MAC SPECIFIC: A breakline in the menu.
	glutAddMenuEntry("Watch movement over time", MENU_MOVEMENT);
	glutAddMenuEntry("-", MENU_QUIT);				// MAC SPECIFIC
	glutAddMenuEntry("Quit", MENU_QUIT);
	glutAttachMenu(GLUT_MIDDLE_BUTTON);
	
	glutMainLoop();
	return 0;
}